==> qtools_profiler-2.2.x-dev.tar.gz: qtools_profiler/modules/qtools_cache_profiler/js/cache-profiler-preview.js <==
/**
 * Inspired by https://www.drupal.org/project/renderviz.
 */
(function (Drupal, $, once, window) {
  const observer = qToolsProfilerPreviewObserver;

  const qToolsCacheProfiler = {
    // Store metadata for all rendered elements on the page.
    cachebleMetadata: {},

    /**
     * Calculate possible cache complexity as log(base 100).
     */
    _calculateComplexity: function (contexts, scales, ignoredContexts) {
      let complexity = 1;
      let matches = [];

      for (let i = 0; i < contexts.length; i++) {
        const context = contexts[i];
        if (ignoredContexts.indexOf(context) !== -1) {
          continue;
        }
        for (let j = 0; j < scales.length; j++) {
          const scale = scales[j];
          const rex = new RegExp(scale[0]);
          if (context.match(rex)) {
            if (scale[1] > 1) {
              complexity = complexity * scale[1];
              matches.push(context + ':' + scale[1]);
            }
            break;
          }
        }
      }

      // Result is capped at 4 and is lx(100).
      return {
        level: Math.max(1, Math.min(4, Math.round(Math.log(complexity) / Math.log(100)))),
        matches: matches.join(',')
      }
    },

    /**
     * Sets element classes based on his metadata.
     */
    _setElementAttributes: function (element, rendererId, final, prebubling, cid, settings) {
      const $element = $(element);

      // If element already processed skip.
      if ($element.data('qtcp-rid')) {
        return;
      }

      // Thi helps significantly reduce scope and
      // exclude system placeholders.
      if ($element.parents('.dialog-off-canvas-main-canvas').length === 0) {
        return;
      }

      // Trick to make 'class' attribute to appear before our data attributes.
      $element.addClass('__qtcp__').removeClass('__qtcp__');

      // For normal elements we only interested in prebubling here, but
      // if element lazy loaded then prebubling info will be taken form
      // placeholder and not from element, so we need to check final metadata.
      const metadata = cid.lazy && final || prebubling;
      let uncacheableReason = '';

      // Case with max-age=1 it's a static cache from QToolsCacheHelper.
      if ((metadata['max-age'] === 0 || metadata['max-age'] === 1) && !$element.hasClass('qtcp-bp-placeholder')) {
        $element.addClass('qtcp-uncacheable');
        uncacheableReason = 'max-age=' + metadata['max-age'];
      }
      else if (cid['cid'] === '') {
        // If cid is empty, item wasn't cached either, so we need to check if
        // it is inside cacheable context.
        if ($element.parent().hasClass('qtcp-bp-placeholder')) {
          $element.addClass('qtcp-uncacheable');
          uncacheableReason = "cid = ''";
        }
      }
      if (cid['e'] > 0.1) {
        console.log('SLOW: ' + cid['e'], rendererId, $element);
      }

      // If element is uncacheable - mark parents as trail.
      if ($element.hasClass('qtcp-uncacheable')) {
        // Report uncacheable elements that are not lazyloaded.
        if ($element.parents('.qtcp-bp-placeholder').length === 0) {
          console.log('UNCACHEABLE: ' + uncacheableReason, rendererId, $element);
        }

        let $parent = $element;
        let level = 0;
        while (
          $parent.length > 0 &&
          level < 10 &&
          !$parent.hasClass('qtcp-block') &&
          !$parent.hasClass('qtcp-bp-placeholder')
          ) {
          $parent = $parent.parent();
          $parent.addClass('qtcp-uncacheable-trail');
          $parent.attr('data-qtcp-uncacheable-trail-level', level);
          level = level + 1;
        }
      }

      // If element is a block we need to set its complexity
      // (if it's not uncacheable completely).
      if ($element.hasClass('qtcp-block') &&
        !$element.hasClass('qtcp-uncacheable-trail') &&
        !$element.hasClass('qtcp-uncacheable')
      ) {
        const info = this._calculateComplexity(final['contexts'], settings.qtools_profiler.contexts_scales, []);
        if (info.level > 0) {
          $element.addClass('qtcp-complexity-' + info.level);
          $element.attr('data-qtcp-complexity-matches', info.matches);
        }
      }

      // Mark element with rendererId.
      $element.attr('data-qtcp-rid', rendererId.trim());
    },

    /**
     * Processes all HTML comments in the CacheableMetadata.
     */
    _processCommentsIntoCacheableMetadata: function (context, settings) {
      // Parse all renderer comments into dataset.
      const comments = $(context).comments(true);
      for (let i = 0; i < comments.length; i++) {
        if (comments[i].textContent.indexOf('RENDERER_START') === 0) {
          const rendererId = comments[i].textContent.split(':')[1].trim();
          const element = comments[i].nextElementSibling;
          if (element) {
            const hash = comments[i + 1].textContent + comments[i + 2].textContent;
            const prebubling = JSON.parse(comments[i + 1].textContent);
            const final = JSON.parse(comments[i + 2].textContent);
            const cid = JSON.parse(comments[i + 3].textContent);
            if (!this.cachebleMetadata[rendererId]) {
              this.cachebleMetadata[rendererId] = {
                hash,
                final,
                prebubling,
                cid,
              }
            }
            else if (this.cachebleMetadata[rendererId].hash !== comments[i + 1].textContent + comments[i + 2].textContent) {
              // Normally this never happen.
              console.log('Metadata collision', rendererId);
            }

            // Add classes to the element based on his metadata.
            this._setElementAttributes(element, rendererId, final, prebubling, cid, settings);
          }
          // @todo is this still the possible case?
          else {
            // Some templates may return empty content, mostly empty fields.
            // For fields there is no need to worry about.
            if (comments[i + 4].textContent.trim() === 'THEME DEBUG' && comments[i + 5].textContent.trim() === "THEME HOOK: 'field'") {
              // Atm do nothing.
            }
            else if ($(comments[i].parentNode).hasClass('qtcp-bp-placeholder')) {
              // Placeholder wrappers can also be empty.
              // Uncomment if more info needed.
              // console.log(comments[i].parentNode);
            }
            else {
              // Uncomment if more info needed.
              console.log(comments[i], rendererId, comments[i].parentNode, 'nextElementSibling not found');
            }
          }
        }
      }
    },

    /**
     * Handles profiler report selection.
     */
    _handleProfilerReportSelection: function (context, settings) {
      const $qtcpReportSelection =  $(once('qtcp-report-selection', 'body'));
      $qtcpReportSelection.each(function () {
        // Subscribe to render cache report selection.
        observer.subscribe({
          'onSelectReport': function (data) {
            $('html').toggleClass('qtcp-active', data.report === 'render_cache');
          },
        });
      });
    },

    /**
     * Drupal attach behavior hook.
     */
    attach: function (context, settings) {
      // Gather info from context.
      this._processCommentsIntoCacheableMetadata(context, settings)

      // Add an ability to inspect elements.
      this._handleProfilerReportSelection(context, settings);
    },

    /**
     * Lookup metadata for given renderId.
     */
    lookup: function (renderId) {
      const $element = $('[data-qtcp-rid="' + renderId + '"]');
      return {
        element: $element,
        final: cachebleMetadata.final,
        prebubling: cachebleMetadata.prebubling,
      };
    },

    /**
     * Finds elements that match max-age criteria.
     */
    findMaxAge: function (maxAgeValue, scopeRenderId) {
      const result = {};
      const that = this;
      Object.keys(this.cachebleMetadata).forEach((renderId) => {
        const cachebleMetadata = that.cachebleMetadata[renderId];

        // Loose matching.
        if (cachebleMetadata.prebubling['max-age'] == maxAgeValue) {
          const $element = $('[data-qtcp-rid="' + renderId + '"]');
          if ($element.length > 0 &&
            (!scopeRenderId || $element.parents('[data-qtcp-rid="' + scopeRenderId + '"]').length > 0)
          ) {
            result[renderId] = {
              element: $element,
              final: cachebleMetadata.final,
              prebubling: cachebleMetadata.prebubling,
            };
          }
        }
      });
      return result;
    },

    /**
     * Finds elements that match cache context criteria.
     */
    findCacheContext: function (cacheContextValue, scopeRenderId) {
      const result = {};
      const that = this;
      Object.keys(this.cachebleMetadata).forEach((renderId) => {
        const cachebleMetadata = that.cachebleMetadata[renderId];

        // Loose matching.
        if (cachebleMetadata.prebubling['contexts'].indexOf(cacheContextValue) !== -1) {
          const $element = $('[data-qtcp-rid="' + renderId + '"]');
          if ($element.length > 0 &&
            (!scopeRenderId || $element.parents('[data-qtcp-rid="' + scopeRenderId + '"]').length > 0)
          ) {
            result[renderId] = {
              element: $element,
              final: cachebleMetadata.final,
              prebubling: cachebleMetadata.prebubling,
            };
          }
        }
      });
      return result;
    },

    /**
     * Finds elements that match cache context criteria.
     */
    findCacheTag: function (cacheTagValue, scopeRenderId) {
      const result = {};
      const that = this;
      Object.keys(this.cachebleMetadata).forEach((renderId) => {
        const cachebleMetadata = that.cachebleMetadata[renderId];

        // Loose matching.
        if (cachebleMetadata.prebubling['tags'].indexOf(cacheTagValue) !== -1) {
          const $element = $('[data-qtcp-rid="' + renderId + '"]');
          if ($element.length > 0 &&
            (!scopeRenderId || $element.parents('[data-qtcp-rid="' + scopeRenderId + '"]').length > 0)
          ) {
            result[renderId] = {
              element: $element,
              final: cachebleMetadata.final,
              prebubling: cachebleMetadata.prebubling,
            };
          }
        }
      });
      return result;
    }
  };

  Drupal.behaviors.qToolsCacheProfiler = qToolsCacheProfiler;
  window._qtcp = qToolsCacheProfiler;

})(Drupal, jQuery, once, window);

==> qtools_profiler-2.3.x-dev.tar.gz: qtools_profiler/modules/qtools_cache_profiler/js/cache-profiler-preview.js <==
/**
 * Inspired by https://www.drupal.org/project/renderviz.
 */
(function (Drupal, $, once, window) {
  const observer = qToolsProfilerPreviewObserver;

  const qToolsCacheProfiler = {
    // Store metadata for all rendered elements on the page.
    cachebleMetadata: {},

    /**
     * Calculate possible cache complexity as log(base 100).
     */
    _calculateComplexity: function (contexts, scales, ignoredContexts) {
      let complexity = 1;
      let matches = [];

      for (let i = 0; i < contexts.length; i++) {
        const context = contexts[i];
        if (ignoredContexts.indexOf(context) !== -1) {
          continue;
        }
        for (let j = 0; j < scales.length; j++) {
          const scale = scales[j];
          const rex = new RegExp(scale[0]);
          if (context.match(rex)) {
            if (scale[1] > 1) {
              complexity = complexity * scale[1];
              matches.push(context + ':' + scale[1]);
            }
            break;
          }
        }
      }

      // Result is capped at 4 and is lx(100).
      return {
        level: Math.max(1, Math.min(4, Math.round(Math.log(complexity) / Math.log(100)))),
        matches: matches.join(',')
      }
    },

    /**
     * Sets element classes based on his metadata.
     */
    _setElementAttributes: function (element, rendererId, final, prebubling, cid, settings) {
      const $element = $(element);

      // If element already processed skip.
      if ($element.data('qtcp-rid')) {
        return;
      }

      // Thi helps significantly reduce scope and
      // exclude system placeholders.
      if ($element.parents('.dialog-off-canvas-main-canvas').length === 0) {
        return;
      }

      // Trick to make 'class' attribute to appear before our data attributes.
      $element.addClass('__qtcp__').removeClass('__qtcp__');

      // For normal elements we only interested in prebubling here, but
      // if element lazy loaded then prebubling info will be taken form
      // placeholder and not from element, so we need to check final metadata.
      const metadata = cid.lazy && final || prebubling;
      let uncacheableReason = '';

      // Case with max-age=1 it's a static cache from QtoolsCacheHelper.
      if ((metadata['max-age'] === 0 || metadata['max-age'] === 1) && !$element.hasClass('qtcp-bp-placeholder')) {
        $element.addClass('qtcp-uncacheable');
        uncacheableReason = 'max-age=' + metadata['max-age'];
      }
      else if (cid['cid'] === '') {
        // If cid is empty, item wasn't cached either, so we need to check if
        // it is inside cacheable context.
        if ($element.parent().hasClass('qtcp-bp-placeholder')) {
          $element.addClass('qtcp-uncacheable');
          uncacheableReason = "cid = ''";
        }
      }
      if (cid['e'] > 0.1) {
        console.log('SLOW: ' + cid['e'], rendererId, $element);
      }

      // If element is uncacheable - mark parents as trail.
      if ($element.hasClass('qtcp-uncacheable')) {
        // Report uncacheable elements that are not lazyloaded.
        if ($element.parents('.qtcp-bp-placeholder').length === 0) {
          console.log('UNCACHEABLE: ' + uncacheableReason, rendererId, $element);
        }

        let $parent = $element;
        let level = 0;
        while (
          $parent.length > 0 &&
          level < 10 &&
          !$parent.hasClass('qtcp-block') &&
          !$parent.hasClass('qtcp-bp-placeholder')
          ) {
          $parent = $parent.parent();
          $parent.addClass('qtcp-uncacheable-trail');
          $parent.attr('data-qtcp-uncacheable-trail-level', level);
          level = level + 1;
        }
      }

      // If element is a block we need to set its complexity
      // (if it's not uncacheable completely).
      if ($element.hasClass('qtcp-block') &&
        !$element.hasClass('qtcp-uncacheable-trail') &&
        !$element.hasClass('qtcp-uncacheable')
      ) {
        const info = this._calculateComplexity(final['contexts'], settings.qtools_profiler.contexts_scales, []);
        if (info.level > 0) {
          $element.addClass('qtcp-complexity-' + info.level);
          $element.attr('data-qtcp-complexity-matches', info.matches);
        }
      }

      // Mark element with rendererId.
      $element.attr('data-qtcp-rid', rendererId.trim());
    },

    /**
     * Processes all HTML comments in the CacheableMetadata.
     */
    _processCommentsIntoCacheableMetadata: function (context, settings) {
      // Parse all renderer comments into dataset.
      const comments = $(context).comments(true);
      for (let i = 0; i < comments.length; i++) {
        if (comments[i].textContent.indexOf('RENDERER_START') === 0) {
          const rendererId = comments[i].textContent.split(':')[1].trim();
          const element = comments[i].nextElementSibling;
          if (element) {
            const hash = comments[i + 1].textContent + comments[i + 2].textContent;
            const prebubling = JSON.parse(comments[i + 1].textContent);
            const final = JSON.parse(comments[i + 2].textContent);
            const cid = JSON.parse(comments[i + 3].textContent);
            if (!this.cachebleMetadata[rendererId]) {
              this.cachebleMetadata[rendererId] = {
                hash,
                final,
                prebubling,
                cid,
              }
            }
            else if (this.cachebleMetadata[rendererId].hash !== comments[i + 1].textContent + comments[i + 2].textContent) {
              // Normally this never happen.
              console.log('Metadata collision', rendererId);
            }

            // Add classes to the element based on his metadata.
            this._setElementAttributes(element, rendererId, final, prebubling, cid, settings);
          }
          // @todo is this still the possible case?
          else {
            // Some templates may return empty content, mostly empty fields.
            // For fields there is no need to worry about.
            if (comments[i + 4].textContent.trim() === 'THEME DEBUG' && comments[i + 5].textContent.trim() === "THEME HOOK: 'field'") {
              // Atm do nothing.
            }
            else if ($(comments[i].parentNode).hasClass('qtcp-bp-placeholder')) {
              // Placeholder wrappers can also be empty.
              // Uncomment if more info needed.
              // console.log(comments[i].parentNode);
            }
            else {
              // Uncomment if more info needed.
              console.log(comments[i], rendererId, comments[i].parentNode, 'nextElementSibling not found');
            }
          }
        }
      }
    },

    /**
     * Handles profiler report selection.
     */
    _handleProfilerReportSelection: function (context, settings) {
      const $qtcpReportSelection =  $(once('qtcp-report-selection', 'body'));
      $qtcpReportSelection.each(function () {
        // Subscribe to render cache report selection.
        observer.subscribe({
          'onSelectReport': function (data) {
            $('html').toggleClass('qtcp-active', data.report === 'render_cache');
          },
        });
      });
    },

    /**
     * Drupal attach behavior hook.
     */
    attach: function (context, settings) {
      // Gather info from context.
      this._processCommentsIntoCacheableMetadata(context, settings)

      // Add an ability to inspect elements.
      this._handleProfilerReportSelection(context, settings);
    },

    /**
     * Lookup metadata for given renderId.
     */
    lookup: function (renderId) {
      const $element = $('[data-qtcp-rid="' + renderId + '"]');
      return {
        element: $element,
        final: cachebleMetadata.final,
        prebubling: cachebleMetadata.prebubling,
      };
    },

    /**
     * Finds elements that match max-age criteria.
     */
    findMaxAge: function (maxAgeValue, scopeRenderId) {
      const result = {};
      const that = this;
      Object.keys(this.cachebleMetadata).forEach((renderId) => {
        const cachebleMetadata = that.cachebleMetadata[renderId];

        // Loose matching.
        if (cachebleMetadata.prebubling['max-age'] == maxAgeValue) {
          const $element = $('[data-qtcp-rid="' + renderId + '"]');
          if ($element.length > 0 &&
            (!scopeRenderId || $element.parents('[data-qtcp-rid="' + scopeRenderId + '"]').length > 0)
          ) {
            result[renderId] = {
              element: $element,
              final: cachebleMetadata.final,
              prebubling: cachebleMetadata.prebubling,
            };
          }
        }
      });
      return result;
    },

    /**
     * Finds elements that match cache context criteria.
     */
    findCacheContext: function (cacheContextValue, scopeRenderId) {
      const result = {};
      const that = this;
      Object.keys(this.cachebleMetadata).forEach((renderId) => {
        const cachebleMetadata = that.cachebleMetadata[renderId];

        // Loose matching.
        if (cachebleMetadata.prebubling['contexts'].indexOf(cacheContextValue) !== -1) {
          const $element = $('[data-qtcp-rid="' + renderId + '"]');
          if ($element.length > 0 &&
            (!scopeRenderId || $element.parents('[data-qtcp-rid="' + scopeRenderId + '"]').length > 0)
          ) {
            result[renderId] = {
              element: $element,
              final: cachebleMetadata.final,
              prebubling: cachebleMetadata.prebubling,
            };
          }
        }
      });
      return result;
    },

    /**
     * Finds elements that match cache context criteria.
     */
    findCacheTag: function (cacheTagValue, scopeRenderId) {
      const result = {};
      const that = this;
      Object.keys(this.cachebleMetadata).forEach((renderId) => {
        const cachebleMetadata = that.cachebleMetadata[renderId];

        // Loose matching.
        if (cachebleMetadata.prebubling['tags'].indexOf(cacheTagValue) !== -1) {
          const $element = $('[data-qtcp-rid="' + renderId + '"]');
          if ($element.length > 0 &&
            (!scopeRenderId || $element.parents('[data-qtcp-rid="' + scopeRenderId + '"]').length > 0)
          ) {
            result[renderId] = {
              element: $element,
              final: cachebleMetadata.final,
              prebubling: cachebleMetadata.prebubling,
            };
          }
        }
      });
      return result;
    }
  };

  Drupal.behaviors.qToolsCacheProfiler = qToolsCacheProfiler;
  window._qtcp = qToolsCacheProfiler;

})(Drupal, jQuery, once, window);
